import trading_api_wrappers as wrappers

from .base import Converter

__all__ = [
    'OpenExchangeRates'
]


class OpenExchangeRates(Converter):
    name = 'OpenExchangeRates'
    slug = 'open-exchange-rates'

    def __init__(self, return_decimal: bool=False, timeout: int=None, retry: bool=None):
        super().__init__(return_decimal)
        app_id = self.credentials['app_id']
        self.client = wrappers.OXR(app_id, timeout, retry)

    def _get_rate(self, currency: str, to: str):
        response = self.client.latest(base=currency.lower(), symbols=[to.lower()])
        rate = response['rates'][to.upper()]
        return rate



import asyncio
import time
from logging import Logger

import trading_bots
from .registry import bots
from ..conf import settings


class BotTask:
    """Class representing a Bot Task."""

    def __init__(self, bot_label: str, config_name: str, logger: Logger=None):
        settings.configure()
        trading_bots.setup()
        self.config_name = config_name
        self.logger = logger
        self.bot_cls = bots.get_bot(bot_label).cls
        self.bot_config = bots.get_config(bot_label, config_name)

    def get_bot_instance(self):
        return self.bot_cls(self.bot_config, self.config_name, self.logger)

    def run_once(self):
        bot_instance = self.get_bot_instance()
        bot_instance.execute()

    def abort(self):
        bot_instance = self.get_bot_instance()
        bot_instance.abort()

    def loop(self, interval: int):
        while True:
            self.run_once()
            time.sleep(interval)

    async def loop_async(self, interval: int):
        while True:
            self.run_once()
            await asyncio.sleep(interval)

    def run_loop(self, interval: int):
        self.loop(interval)

    def run_loop_async(self, interval: int):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.loop_async(interval))
        loop.close()



import os
import unittest

from trading_bots.core.storage import *


class JSONStorageTest(unittest.TestCase):

    def setUp(self):
        self.filename = 'test.json'
        self.types_dict = {
            'str': 'a',
            'int': 1,
            'float': 1.0,
            'bool': True,
            'list': ['a', 1, 1.0, True],
            'test': 'test',
        }
        self.sample_json = {
            **self.types_dict,
            'dict': {
                **self.types_dict,
            },
        }
        with open(self.filename, 'w') as outfile:
            j.dump(self.sample_json, outfile)

        self.store = JSONStore(self.filename)

    def tearDown(self):
        try:
            os.remove(self.filename)
        except OSError:
            pass

    def test_instantiate_bot(self):
        self.assertIsInstance(self.store, Store)

    # GET --------------------------------------------------------------------
    def test_get(self):
        for name, value in self.sample_json.items():
            stored_value = self.store.get(name)
            self.assertEqual(stored_value, value)

    def test_get_none(self):
        none_value = self.store.get('foo')
        self.assertIsNone(none_value)

    def test_hget(self):
        for key, value in self.sample_json['dict'].items():
            stored_value = self.store.hget('dict', key)
            self.assertEqual(stored_value, value)

    def test_hget_none(self):
        none_value = self.store.hget('dict', 'foo')
        self.assertIsNone(none_value)

    # GET --------------------------------------------------------------------
    def test_set_add(self):
        self.store.set('foo', self.types_dict)
        stored_value = self.store.get('foo')
        self.assertDictEqual(stored_value, self.types_dict)

    def test_set_replace(self):
        self.store.set('test', self.types_dict)
        stored_value = self.store.get('test')
        self.assertDictEqual(stored_value, self.types_dict)

    def test_hset_add(self):
        self.store.hset('foo', 'bar', self.types_dict)
        stored_value = self.store.hget('foo', 'bar')
        self.assertDictEqual(stored_value, self.types_dict)

    def test_hset_replace(self):
        self.store.hset('test', 'foo', self.types_dict)
        stored_value = self.store.hget('test', 'foo')
        self.assertDictEqual(stored_value, self.types_dict)

    def test_hset_update(self):
        self.store.hset('dict', 'test', 'foo')
        stored_value = self.store.get('dict')
        self.assertDictEqual(stored_value, {**self.types_dict, 'test': 'foo'})

    def test_set_add(self):
        self.store.set('foo', self.types_dict)
        stored_value = self.store.get('foo')
        self.assertDictEqual(stored_value, self.types_dict)

    # DEL --------------------------------------------------------------------
    def test_delete(self):
        self.store.delete('test')
        value = self.store.get('test')
        self.assertIsNone(value)

    def test_delete_none(self):
        self.store.delete('foo')
        value = self.store.get('foo')
        self.assertIsNone(value)

    def test_hdel(self):
        self.store.hdel('dict', 'test')
        value = self.store.hget('dict', 'test')
        self.assertIsNone(value)

    def test_hdel_none(self):
        self.store.hdel('foo', 'bar')
        value = self.store.hget('foo', 'bar')
        self.assertIsNone(value)



import trading_api_wrappers as wrappers

from .base import Converter

__all__ = [
    'CurrencyLayer'
]


class CurrencyLayer(Converter):
    name = 'Currencylayer'
    slug = 'currencylayer'

    def __init__(self, return_decimal: bool=False, timeout: int=None, retry: bool=None):
        super().__init__(return_decimal)
        access_key = self.credentials['access_key']
        self.client = wrappers.CurrencyLayer(access_key, timeout, retry)

    def _get_rate(self, currency: str, to: str):
        market = (currency + to).upper()
        response = self.client.live_rates(base=currency.lower(), currencies=[to.lower()])
        rate = response['quotes'][market]
        return rate



from trading_api_wrappers import Bitfinex

from .base import *

__all__ = [
    'BitfinexPublic',
    'BitfinexAuth',
    'BitfinexMarket',
    'BitfinexWallet',
    'BitfinexTrading',
]

DEFAULT_WALLET_TYPE = 'exchange'


class BitfinexPublic(BaseClient):
    name = 'Bitfinex'

    def _client(self):
        return Bitfinex.Public(timeout=self.timeout)


class BitfinexAuth(BaseClient):
    name = 'Bitfinex'

    def _client(self):
        key = self.credentials['key']
        secret = self.credentials['secret']
        return Bitfinex.Auth(key, secret, timeout=self.timeout)


class BitfinexMarket(MarketClient, BitfinexPublic):

    def _ticker(self):
        return self.client.ticker(symbol=self.market_id)

    def _order_book(self, side: Side=None):
        order_book = self.client.order_book(symbol=self.market_id)
        bids, asks = order_book['bids'], order_book['asks']
        if side:
            return bids if side == Side.BUY else asks
        return OrderBook(bids=bids, asks=asks)

    def _order_book_entry_amount(self, order):
        return float(order['amount'])

    def _order_book_entry_price(self, order):
        return float(order['price'])


class BitfinexWallet(WalletClient, BitfinexAuth):
    withdrawal_fees = {
        'BCH': 0.0005,
        'BTC': 0.0005,
        'ETH': 0.01,
        'LTC': 0.01,
    }
    method_mapping = {
        'BCH': 'bcash',
        'BTC': 'bitcoin',
        'ETH': 'ethereum',
        'LTC': 'litecoin',
    }

    def __init__(self, currency: str, client=None, dry_run: bool=False, timeout: int=None, logger=None,
                 wallet_type: str=DEFAULT_WALLET_TYPE):
        super().__init__(currency, client, dry_run, timeout, logger)
        self.wallet_type = wallet_type

    def _balance(self, currency, available_only=False):
        balances = self.client.balances()
        currency = currency.lower()
        balance = next(b for b in balances if b['currency'] == currency
                       and b['type'] == self.wallet_type)
        if available_only:
            return float(balance['available'])
        return float(balance['amount'])

    @staticmethod
    def _filter_state(items, state: str=None):
        if state:
            items = [i for i in items if i['status'] == state]
        return items

    def _deposits(self, currency: str):
        deposits = self.client.movements(currency)
        return [d for d in deposits if d['type'] == 'DEPOSIT']

    def _withdrawals(self, currency: str):
        withdrawals = self.client.movements(currency)
        return [w for w in withdrawals if w['type'] == 'WITHDRAWAL']

    def _withdraw(self, currency: str, amount: float, address: str, subtract_fee: bool=False):
        method = self.method_mapping[currency]
        if subtract_fee:
            fee = self.withdrawal_fees[currency]
            amount -= fee
        return self.client.withdraw(method, 'exchange', amount, address)


class BitfinexTrading(TradingClient, BitfinexAuth, BitfinexMarket):
    wallet_client = BitfinexWallet
    has_margin_trading = True
    min_order_amount_mapping = {
        'BCH': 0.02,
        'BTC': 0.002,
        'ETH': 0.04,
        'LTC': 0.02,
    }
    order_type_mapping = {
        OrderType.MARKET: 'exchange market',
        OrderType.LIMIT: 'exchange limit',
    }

    def _open_orders(self):
        orders = self.client.active_orders()
        return [o for o in orders if o['is_live'] is True]

    def _order_amount(self, order):
        return order['remaining_amount']

    def _cancel_order(self, order):
        return self.client.delete_order(order['id'])

    def _order_details(self, order_id: int):
        return self.client.status_order(order_id)

    def _open_positions(self):
        positions = self.client.active_positions()
        return [p for p in positions if p['symbol'] == self.market_id and p['status'] in ['ACTIVE']]

    @staticmethod
    def _order_price(order_type: OrderType, price: float):
        return price if order_type == OrderType.LIMIT else 1  # Bitfinex API doc: 'Use random number for market orders'

    def _place_order(self, side: Side, o_type: OrderType, amount: float, price: float=None):
        price = self._order_price(o_type, price)
        order_type = self.order_type_mapping[o_type]
        return self.client.place_order(amount, price, side.value, order_type, self.market_id)

    def _open_position(self, side: Side, p_type: OrderType, amount: float, price: float=None, leverage: float=None):
        price = self._order_price(p_type, price)
        return self.client.place_order(amount, price, side.value, p_type.value, self.market_id)



import simple_settings


class LazySettings(simple_settings.LazySettings):
    DEFAULT_SETTINGS = 'trading_bots.conf.global_settings'
    ENVIRON_KEYS = ('settings', 'SETTINGS')
    COMMAND_LINE_ARGS = ('--settings',)

    def __init__(self):
        settings_value = self._get_settings_value()
        settings_list = settings_value.split(',') if settings_value else {}
        super().__init__(self.DEFAULT_SETTINGS, *settings_list)


settings = LazySettings()



import json as j
import pickle as p
from logging import Logger

from .logging import get_logger
from .utils import load_class_by_name

try:
    import redis
except ImportError:
    pass


def get_store(logger: Logger=None):
    """Get and configure the storage backend"""
    from trading_bots.conf import settings
    store_settings = settings.storage
    store = store_settings.get('name', 'json')
    if store == 'json':
        store = 'trading_bots.core.storage.JSONStore'
    elif store == 'redis':
        store = 'trading_bots.core.storage.RedisStore'
    store_cls = load_class_by_name(store)
    kwargs = store_cls.configure(store_settings)
    return store_cls(logger=logger, **kwargs)


class BaseStore:
    name = ''

    def __init__(self, logger: Logger=None):
        assert self.name, 'A name name must be defined!'
        self.log = logger or get_logger(__name__)

    @classmethod
    def configure(cls, settings):
        """Configure the storage method with app settings"""
        return {}


class Store(BaseStore):
    serializers = {
        'json': j,
        'pickle': p,
    }

    # GET --------------------------------------------------------------------
    def __get(self, method, *path, cast=None, serializer=None, **kwargs):
        path_str = ' '.join(path)
        self.log.debug(f'Get {path_str} from {self.name}')
        try:
            value = method(*path, **kwargs)
            if serializer:
                if isinstance(serializer, str):
                    serializer = self.serializers[serializer]
                return serializer.loads(value)
            if cast:
                return cast(value)
            return value
        except KeyError:
            self.log.warning(f'{path_str} not found on {self.name}')
            return None
        except Exception:
            self.log.exception(f'Failed to get {path_str} from {self.name}')
            raise

    def _get(self, name: str, **kwargs):
        raise NotImplementedError

    def get(self, name: str, cast=None, serializer=None, **kwargs):
        return self.__get(self._get, name, cast=cast, serializer=serializer, **kwargs)

    def _hget(self, name: str, key: str, **kwargs):
        raise NotImplementedError

    def hget(self, name: str, key: str, cast=None, serializer=None, **kwargs):
        return self.__get(self._hget, name, key, cast=cast, serializer=serializer, **kwargs)

    # SET --------------------------------------------------------------------
    def __set(self, method, *path, value, serializer=None, **kwargs):
        path_str = ' '.join(path)
        self.log.debug(f'Set {path_str} on {self.name}')
        try:
            if serializer:
                if isinstance(serializer, str):
                    serializer = self.serializers[serializer]
                value = serializer.dumps(value)
            method(*path, value=value, **kwargs)
        except Exception:
            self.log.exception(f'Failed to set {path_str} on {self.name}')
            raise

    def _set(self, name: str, value, **kwargs):
        raise NotImplementedError

    def set(self, name: str, value, serializer=None, **kwargs):
        return self.__set(self._set, name, value=value, serializer=serializer, **kwargs)

    def _hset(self, name: str, key: str, value, **kwargs):
        raise NotImplementedError

    def hset(self, name: str, key: str, value, serializer=None, **kwargs):
        return self.__set(self._hset, name, key, value=value, serializer=serializer, **kwargs)

    # DELETE -----------------------------------------------------------------
    def __delete(self, method, *path, **kwargs):
        path_str = ' '.join(path)
        self.log.debug(f'Delete {path_str} from {self.name}')
        try:
            method(*path, **kwargs)
        except KeyError:
            self.log.warning(f'{path_str} not found on {self.name}')
            pass
        except Exception:
            self.log.exception(f'Failed to delete {path_str} from {self.name}')
            raise

    def _delete(self, name: str, **kwargs):
        raise NotImplementedError

    def delete(self, name: str, **kwargs):
        return self.__delete(self._delete, name, **kwargs)

    def _hdel(self, name: str, key: str, **kwargs):
        raise NotImplementedError

    def hdel(self, name: str, key: str, **kwargs):
        return self.__delete(self._hdel, name, key, **kwargs)


class JSONStore(Store):
    name = 'JSON File'
    filename = 'store.json'

    def __init__(self, filename: str=None, logger: Logger=None):
        super().__init__(logger)
        if filename is not None:
            self.filename = filename

    @classmethod
    def configure(cls, settings):
        kwargs = super().configure(settings)
        kwargs['filename'] = settings.get('filename')
        return kwargs

    def _read(self):
        try:
            with open(self.filename) as data_file:
                return j.load(data_file)
        except FileNotFoundError:
            self.log.warning(f'File not found! ({self.name})')
            return {}

    def _write(self, value):
        try:
            with open(self.filename, 'w') as outfile:
                j.dump(value, outfile)
        except Exception:
            self.log.exception(f'Failed to write to {self.name}!')
            raise

    def _get(self, name: str, **kwargs):
        data = self._read()
        return data[name]

    def _hget(self, name: str, key: str, **kwargs):
        return self._get(name)[key]

    def _set(self, name: str, value, **kwargs):
        data = self._read()
        data[name] = value
        return self._write(data)

    def _hset(self, name, key, value, **kwargs):
        data = self._read()
        old = data.get(name, {})
        if not isinstance(old, dict):
            old = {}
        data[name] = {**old, key: value}
        return self._write(data)

    def _delete(self, name: str, **kwargs):
        data = self._read() or {}
        data.pop(name)
        return self._write(data)

    def _hdel(self, name: str, key: str, **kwargs):
        data = self._get(name)
        data.pop(key)
        return self._write(data)


class RedisStore(Store):
    name = 'Redis'

    def __init__(self, url: str, logger: Logger=None):
        super().__init__(logger)
        self.r = redis.StrictRedis.from_url(url)

    @classmethod
    def configure(cls, settings):
        kwargs = super().configure(settings)
        kwargs['url'] = settings.get('url')
        return kwargs

    def _get(self, name: str, **kwargs):
        value = self.r.get(name)
        if value is None:
            raise KeyError
        return value

    def _hget(self, name: str, key: str, **kwargs):
        value = self.r.hget(name, key)
        if value is None:
            raise KeyError
        return value

    def _set(self, name: str, value, **kwargs):
        return self.r.set(name, value)

    def _hset(self, name, key, value, **kwargs):
        return self.r.hset(name, key, value)

    def _delete(self, name: str, **kwargs):
        deleted = self.r.delete(name)
        if not deleted:
            raise KeyError
        return deleted

    def _hdel(self, name: str, key: str, **kwargs):
        deleted = self.r.hdel(name, key)
        if not deleted:
            raise KeyError
        return deleted



from decimal import Decimal

from trading_bots.conf import settings
from trading_bots.core.logging import get_logger

__all__ = [
    'ConverterRateError',
    'ConverterValidationError',
    'Converter'
]

logger = get_logger(__name__)
Number = (float, Decimal)


class ConverterRateError(Exception):
    def __init__(self, converter):
        logger.exception(f'{converter} failed to get rate!')
        super().__init__()


class ConverterValidationError(Exception):
    def __init__(self, converter, rate):
        logger.exception(f'{converter} rate is invalid!: {rate}')
        super().__init__()


class Converter(object):
    name = ''
    slug = ''

    def __init__(self, return_decimal: bool=False, **kwargs):
        assert self.name, 'A converter must have a name!'
        self.credentials = settings.credentials.get(self.name)
        self.return_decimal = return_decimal

    def _format_number(self, value: (str, Number)) -> Number:
        if self.return_decimal:
            if not isinstance(value, Decimal):
                return Decimal(value)
        return float(value)

    def _get_rate(self, currency: str, to: str) -> (str, Number):
        raise NotImplementedError

    def get_rate_for(self, currency: str, to: str, reverse: bool=False) -> Number:
        """Get current market rate for currency"""

        # Return 1 when currencies match
        if currency.upper() == to.upper():
            return self._format_number('1.0')

        # Set base and quote currencies
        base, quote = currency, to
        if reverse:
            base, quote = to, currency

        try:  # Get rate from source
            rate = self._get_rate(base, quote)
        except Exception as e:
            raise ConverterRateError(self.name) from e

        # Convert rate to number
        rate = self._format_number(rate)

        try:  # Validate rate value
            assert isinstance(rate, Number)
            assert rate > 0
        except AssertionError as e:
            raise ConverterValidationError(self.name, rate) from e

        # Return market rate
        if reverse:
            return self._format_number('1.0') / rate
        return rate

    def convert(self, amount: Number, currency: str, to: str, reverse: bool=False) -> Number:
        """Convert amount to another currency"""
        rate = self.get_rate_for(currency, to, reverse)
        if self.return_decimal:
            amount = Decimal(amount)
        return amount * rate



import trading_api_wrappers as wrappers

from .base import Converter

__all__ = [
    'CoinMarketCap'
]


class CoinMarketCap(Converter):
    name = 'CoinMarketCap'
    slug = 'coinmarketcap'

    def __init__(self, return_decimal: bool=False, timeout: int=None, retry: bool=None):
        super().__init__(return_decimal)
        self.client = wrappers.CoinMarketCap(timeout, retry)

    def _get_rate(self, currency: str, to: str):
        rate = self.client.price(currency, convert=to)
        return rate



from trading_bots.bots import Bot
from trading_bots.contrib.clients import BudaTrading
from trading_bots.contrib.clients import Market, Side
from trading_bots.utils import truncate_to


class RelativeOrders(Bot):
    label = 'RelativeOrders'

    def _setup(self, config):
        # Set market
        self.market = Market(config['market'])
        # Set buda trading client
        self.buda = BudaTrading(
            self.market, dry_run=self.dry_run, timeout=self.timeout, logger=self.log, store=self.store)

    def _algorithm(self):
        # PREPARE ORDER PRICES
        # Get middle price
        max_bid, min_ask = self.buda.get_spread_details()
        middle_price = (max_bid + min_ask) / 2
        self.log.info(f'Ticker prices   | Bid: {max_bid} | Ask: {min_ask} | Middle: {middle_price}')
        # Offset prices from middle using configured price multipliers
        prices_config = self.config['prices']
        price_buy = truncate_to(middle_price * prices_config['buy_multiplier'], currency=self.market.quote)
        price_sell = truncate_to(middle_price * prices_config['sell_multiplier'], currency=self.market.quote)
        self.log.info(f'Relative prices | Buy: {price_buy} | Sell: {price_sell}')

        # PREPARE ORDER AMOUNTS
        # Cancel open orders to get correct available amounts
        self.log.info('Closing open orders')
        self.buda.cancel_orders()
        # Fetch available amounts
        available_base = self.buda.wallets.base.get_available()
        available_quote = self.buda.wallets.quote.get_available()
        # Adjust amounts to max in config
        amounts_config = self.config['amounts']
        amount_base = min(amounts_config['max_base'], available_base)
        amount_quote = min(amounts_config['max_quote'], available_quote)
        # Get order buy and sell amounts
        # *quote amount must be converted to base
        amount_buy = truncate_to(amount_quote / price_buy, currency=self.market.base)
        amount_sell = truncate_to(amount_base, currency=self.market.base)
        self.log.info(f'Amounts | Buy {amount_buy} {self.market.quote} | Sell {amount_sell} {self.market.base}')

        # PLACE ORDERS
        self.log.info('Starting order deployment')
        self.buda.place_limit_order(side=Side.BUY, amount=amount_buy, price=price_buy)
        self.buda.place_limit_order(side=Side.SELL, amount=amount_sell, price=price_sell)

    def _abort(self):
        self.log.error('Aborting strategy, cancelling all orders')
        try:
            self.buda.cancel_orders()
        except Exception:
            self.log.critical(f'Failed!, some orders might not be cancelled')
            raise
        else:
            self.log.info(f'All open orders were cancelled')
